"""
Module with functions for processing an input datafile into a PyTorch-compatible
format

.............................................................................
idptools-parrot was developed by the Holehouse lab
     Original release ---- 2020

Question/comments/concerns? Raise an issue on github:
https://github.com/idptools/parrot

Licensed under the MIT license. 
"""

import math

import numpy as np
import torch 
import torch.nn as nn
from torch.utils.data import Dataset

from parrot import encode_sequence

def parse_file(tsvfile, datatype, problem_type, num_classes, excludeSeqID=False):
	"""Parse a datafile containing sequences and values.

	Each line of of the input tsv file contains a sequence of amino acids, a value
	(or values) corresponding to that sequence, and an optional sequence ID. This 
	file will be parsed into a more convenient list of lists. 

	If excludeSeqID is False, then the format of each line in the file should be:
	<seqID> <sequence> <value(s)>

	If excludeSeqID is True, then the format of each line in the file should be:
	<sequence> <value(s)>

	`value(s)` will either be a single number if `datatype` is 'sequence' or a
	len(sequence) series of whitespace-separated numbers if it is 'residues'.

	If `problem_type` is 'regression', then each value can be any real number. But
	if it is 'classification' then each value should be an integer in the range
	[0-N] where N is the number of classes.

	Parameters
	----------
	tsvfile : str
		Path to a whitespace-separated datafile
	datatype : str
		Description of the format of the values in `tsvfile`. Providing a string
		other than 'sequence' or 'residues' will produce unintended behavior.
	problem_type : str
		Description of the machine-learning task. Providing a string other than
		'regression' or 'classification' will produce unintended behavior.
	excludeSeqID : bool, optional
		Boolean indicating whether or not each line in `tsvfile` has a sequence ID
		(default is False)

	Returns
	-------
	list of lists
		A list representing the entire `tsvfile`. Each inner list corresponds to a
		single line in the file and has the format [seqID, sequence, values].
	"""

	with open(tsvfile) as f:
		lines = [line.rstrip().split() for line in f]

	# Add a dummy seqID if none are provided
	if excludeSeqID:
		for line in lines:
			line.insert(0, '')

	try:
		if datatype == 'residues':		# A value for each residue in a sequence
			data = [[x[0], x[1], np.array( x[2:], dtype=np.float)] for x in lines]
		elif datatype == 'sequence':	# A single value per sequence
			data = [[x[0], x[1], float(x[2])] for x in lines]
		else:
			raise ValueError('Invalid datatype. Must be "residues" or "sequence".')
	except:
		raise Exception("Input data is not correctly formatted for datatype '%s'" % datatype)

	if problem_type == 'classification':
		if datatype == 'sequence':
			for sample in data:
				sample[2] = int(sample[2])

				# Validate that all of the class labels are valid
				if sample[2] >= num_classes or sample[2] < 0:
					raise ValueError("Invalid class label: %s" % sample[0])

		else:
			for sample in data:
				sample[2] = list(map(int, sample[2]))
				test = np.array(sample[2])

				if np.any(test < 0) or np.any(test >= num_classes):
					raise ValueError("Invalid class label: %s" % sample[0])
				elif len(test) != len(sample[1]):
					raise Exception(
					'Input not properly formatted. Number of values must be equal to sequence length: %s' % sample[0])
	return data

class SequenceDataset(Dataset):
	"""A PyTorch-compatible dataset containing sequences and values

	Stores a collection of sequences as tensors along with their corresponding
	target values. This class is designed to be provided to PyTorch Dataloaders.


	Attributes
	----------
	data : list of lists
		Each inner list represents a single sequence in the dataset and should
		have the format: [seqID, sequence, value(s)]
	encoding_scheme : str
		Description of how an amino acid sequence should be encoded as a numeric 
		vector. Providing a string other than 'onehot', 'biophysics', or 'user' 
		will produce unintended consequences.
	encoder: UserEncoder object, optional
		If encoding_scheme is 'user', encoder should be a UserEncoder object
		that can convert amino acid sequences to numeric vectors. If
		encoding_scheme is not 'user', use None.
	"""

	def __init__(self, data, subset=np.array([]), encoding_scheme='onehot', 
															encoder=None):
		"""
		Parameters
		----------
		data : list of lists
			Each inner list represents a single sequence in the dataset and should
			have the format: [seqID, sequence, value(s)]
		subset : numpy int array, optional
			Array containing the indices of `data` that are to be a part of this
			dataset. Providing an empty array causes all of `data` to be integrated
			into the dataset (default is empty array).
		encoding_scheme : str
			Description of how an amino acid sequence should be encoded as a numeric 
			vector. Providing a string other than 'onehot', 'biophysics', or 'user' 
			will produce unintended consequences.
		encoder: UserEncoder object, optional
			If encoding_scheme is 'user', encoder should be a UserEncoder object
			that can convert amino acid sequences to numeric vectors. If
			encoding_scheme is not 'user', use None.
		"""

		self.encoding_scheme = encoding_scheme
		self.encoder = encoder

		if len(subset) == 0:
			self.data = data
		else:
			all_data = data
			self.data = [all_data[x] for x in subset]

	def __len__(self):
		"""Get the number of sequences in this dataset."""

		return len(self.data)

	def __getitem__(self, idx):
		"""Get the sequence and values of a specified index.

		Converts a string sequence to an encoded numeric vector.

		Parameters
		----------
		idx : int, optional
			Index of the desired sequence

		Returns
		-------
		tuple
			a tuple of a name, sequence vector and its corresponding values
		"""

		if torch.is_tensor(idx):
			idx = idx.tolist()

		if self.encoding_scheme == 'onehot':
			sequence_vector = encode_sequence.one_hot(self.data[idx][1])
		elif self.encoding_scheme == 'biophysics':
			sequence_vector = encode_sequence.biophysics(self.data[idx][1])
		elif self.encoding_scheme == 'user':
			sequence_vector = self.encoder.encode(self.data[idx][1])

		name = self.data[idx][0]
		value = self.data[idx][2]
		sample = (name, sequence_vector, value)

		return sample


def seq_class_collate(batch):
	"""Collates sequences and their values into a batch

	Transforms a collection of tuples of sequence vectors and values into a single
	tuple by stacking along a newly-created batch dimension. This function is 
	specifically designed for classification problems with sequence-mapped data.

	Parameters
	----------
	batch : list
		A list of tuples of the form (sequence_vector, target_value(s))

	Returns
	-------
	tuple
		a tuple with concatenated names, sequence_vectors and target_values
	"""

	names = [item[0] for item in batch]
	orig_seq_vectors = [item[1] for item in batch]
	orig_targets = [item[2] for item in batch]

	longest_seq = len(max(orig_seq_vectors, key=lambda x: len(x)))

	padded_seq_vectors = np.zeros([len(orig_seq_vectors), longest_seq, len(orig_seq_vectors[0][0])])

	for i,j in enumerate(orig_seq_vectors):
		padded_seq_vectors[i][0:len(j)] = j

	padded_seq_vectors = torch.FloatTensor(padded_seq_vectors)
	targets = torch.LongTensor(orig_targets)

	return (names, padded_seq_vectors, targets)


def seq_regress_collate(batch):
	"""Collates sequences and their values into a batch

	Transforms a collection of tuples of sequence vectors and values into a single
	tuple by stacking along a newly-created batch dimension. This function is 
	specifically designed for regression problems with sequence-mapped data.

	Parameters
	----------
	batch : list
		A list of tuples of the form (sequence_vector, target_value(s))

	Returns
	-------
	tuple
		a tuple with concatenated names, sequence_vectors and target_value
	"""

	names = [item[0] for item in batch]
	orig_seq_vectors = [item[1] for item in batch]
	orig_targets = [[item[2]]for item in batch]

	longest_seq = len(max(orig_seq_vectors, key=lambda x: len(x)))

	padded_seq_vectors = np.zeros([len(orig_seq_vectors), longest_seq, len(orig_seq_vectors[0][0])])

	for i,j in enumerate(orig_seq_vectors):
		padded_seq_vectors[i][0:len(j)] = j

	padded_seq_vectors = torch.FloatTensor(padded_seq_vectors)
	targets = torch.FloatTensor(orig_targets)

	return (names, padded_seq_vectors, targets)

def res_class_collate(batch):
	"""Collates sequences and their values into a batch

	Transforms a collection of tuples of sequence vectors and values into a single
	tuple by stacking along a newly-created batch dimension. This function is 
	specifically designed for classification problems with residue-mapped data. To
	account for sequences with different lengths, all sequence vectors are zero-
	padded to the length of the longest sequence in the batch

	Parameters
	----------
	batch : list
		A list of tuples of the form (sequence_vector, target_value(s))

	Returns
	-------
	tuple
		a tuple with concatenated names, sequence_vectors and target_values
	"""

	names = [item[0] for item in batch]
	orig_seq_vectors = [item[1] for item in batch]
	orig_targets = [item[2] for item in batch]

	longest_seq = len(max(orig_seq_vectors, key=lambda x: len(x)))

	padded_seq_vectors = np.zeros([len(orig_seq_vectors), longest_seq, len(orig_seq_vectors[0][0])])
	padded_targets = np.zeros([len(orig_targets), longest_seq])

	for i,j in enumerate(orig_seq_vectors):
		padded_seq_vectors[i][0:len(j)] = j

	for i,j in enumerate(orig_targets):
		padded_targets[i][0:len(j)] = j

	padded_seq_vectors = torch.FloatTensor(padded_seq_vectors)
	padded_targets = torch.LongTensor(padded_targets)

	return (names, padded_seq_vectors, padded_targets)

def res_regress_collate(batch):
	"""Collates sequences and their values into a batch

	Transforms a collection of tuples of sequence vectors and values into a single
	tuple by stacking along a newly-created batch dimension. This function is 
	specifically designed for regression problems with residue-mapped data. To
	account for sequences with different lengths, all sequence vectors are zero-
	padded to the length of the longest sequence in the batch

	Parameters
	----------
	batch : list
		A list of tuples of the form (sequence_vector, target_value(s))

	Returns
	-------
	tuple
		a tuple with concatenated names, sequence_vectors and target_values
	"""

	names = [item[0] for item in batch]
	orig_seq_vectors = [item[1] for item in batch]
	orig_targets = [item[2] for item in batch]

	longest_seq = len(max(orig_seq_vectors, key=lambda x: len(x)))

	padded_seq_vectors = np.zeros([len(orig_seq_vectors), longest_seq, len(orig_seq_vectors[0][0])])
	padded_targets = np.zeros([len(orig_targets), longest_seq])

	for i,j in enumerate(orig_seq_vectors):
		padded_seq_vectors[i][0:len(j)] = j

	for i,j in enumerate(orig_targets):
		padded_targets[i][0:len(j)] = j

	padded_seq_vectors = torch.FloatTensor(padded_seq_vectors)
	padded_targets = torch.FloatTensor(padded_targets).view(
			(len(padded_targets), len(padded_targets[0]), 1))

	return (names, padded_seq_vectors, padded_targets)


def vector_split(v, fraction):
	"""Split a vector randomly by a specified proportion

	Randomly divide the values of a vector into two, non-overlapping smaller 
	vectors. The proportions of the two vectors will be `fraction` and 
	(1 - `fraction`).

	Parameters
	----------
	v : numpy array
		The vector to divide
	fraction : float
		Size proportion for the returned vectors. Should be in the range [0-1].

	Returns
	-------
	numpy array
		a subset of `v` of length `fraction` * len(v) (rounding up)
	numpy array
		a subset of `v` of length (1-`fraction`) * len(v). 
	"""

	segment1 = np.random.choice(v, size=math.ceil(fraction*len(v)), replace=False)
	segment1.sort()
	segment2 = np.setdiff1d(v, segment1, assume_unique=True)
	return segment1, segment2

def read_split_file(split_file):
	"""Read in a split_file

	Parameters
	----------
	split_file : str
		Path to a whitespace-separated splitfile

	Returns
	-------
	numpy int array
		an array of the indices for the training set samples
	numpy int array
		an array of the indices for the validation set samples
	numpy int array
		an array of the indices for the testing set samples
	"""

	with open(split_file) as f:
		lines = [line.rstrip().split() for line in f]
		training_samples = np.array([int(i) for i in lines[0]])
		val_samples = np.array([int(i) for i in lines[1]])
		test_samples = np.array([int(i) for i in lines[2]])
	return training_samples, val_samples, test_samples

def split_data(data_file, datatype, problem_type, num_classes, 
			excludeSeqID=False, split_file=None, encoding_scheme='onehot', 
			encoder=None, percent_val=0.15, percent_test=0.15):
	"""Divide a datafile into training, validation, and test datasets

	Takes in a datafile and specification of the data format and the machine
	learning problem, and returns PyTorch-compatible Dataset objects for
	the training, validation and test sets of the data. The user may optionally
	specify how the dataset should be split into these subsets, as well as how
	protein sequences should be encoded as numeric vectors.

	Parameters
	----------
	data_file : str
		Path to the datafile containing sequences and corresponding values
	datatype : str
		Format of the values within `data_file`. Should be 'sequence' if the
		`data_file` contains a single value per sequence, or 'residues' if
		it contains a value for each residue per sequence.
	problem_type : str
		The machine learning task to be addressed. Should be either 'regression'
		or 'classification'.
	excludeSeqID : bool, optional
		Flag that indicates how `data_file` is formatted. If False (default),
		then each line in the file should begin with a column containing a
		sequence ID. If True, then the datafile will not have this ID column,
		and will begin with the protein sequence.
	split_file : str, optional
		Path to a file containing information on how to divide the data into
		training, validation and test datasets. Default is None, which will
		cause the data to be divided randomly, with proportions based on
		`percent_val` and `percent_test`. If `split_file` is provided it must
		contain 3 lines in the file, corresponding to the training, validation
		and test sets. Each line should have whitespace-separated integer indices
		which correspond to lines in `data_file`.
	encoding_scheme : str, optional
		The method to be used for encoding protein sequences as numeric vectors.
		Currently 'onehot' and 'biophysics' are implemented (default is 'onehot').
	encoder: UserEncoder object, optional
		If encoding_scheme is 'user', encoder should be a UserEncoder object
		that can convert amino acid sequences to numeric vectors. If
		encoding_scheme is not 'user', use None.
	percent_val : float, optional
		If `split_file` is not provided, the fraction of the data that should be
		randomly assigned to the validation set. Should be in the range [0-1]
		(default is 0.15).
	percent_test : float, optional
		If `split_file` is not provided, the fraction of the data that should be
		randomly assigned to the test set. Should be in the range [0-1] (default
		is 0.15). The proportion of the training set will be calculated by the
		difference between 1 and the sum of `percent_val` and `percent_train`, so
		these should not sum to be greater than 1.

	Returns
	-------
	SequenceDataset object
		a dataset containing the training set sequences and values
	SequenceDataset object
		a dataset containing the validation set sequences and values
	SequenceDataset object
		a dataset containing the test set sequences and values
	"""

	data = parse_file(data_file, datatype, problem_type, num_classes, excludeSeqID=excludeSeqID)
	num_samples = len(data)

	if split_file == None:
		percent_train = 1 - percent_val - percent_test

		all_samples = np.arange(num_samples)
		training_samples, val_test_samples = vector_split(all_samples, percent_train)

		# Repeat procedure to split val and test sets
		val_test_fraction = percent_val / (percent_val + percent_test)
		val_samples, test_samples = vector_split(val_test_samples, val_test_fraction)

		# Generate datasets using these random partitions
		train_set = SequenceDataset(data=data, subset=training_samples,
						encoding_scheme=encoding_scheme, encoder=encoder)
		val_set = SequenceDataset(data=data, subset=val_samples,
						encoding_scheme=encoding_scheme, encoder=encoder)
		test_set = SequenceDataset(data=data, subset=test_samples,
						encoding_scheme=encoding_scheme, encoder=encoder)

	else:
		training_samples, val_samples, test_samples = read_split_file(split_file)

		# Generate datasets using the provided partitions
		train_set = SequenceDataset(data=data, subset=training_samples,
						encoding_scheme=encoding_scheme, encoder=encoder)
		val_set = SequenceDataset(data=data, subset=val_samples,
						encoding_scheme=encoding_scheme, encoder=encoder)
		test_set = SequenceDataset(data=data, subset=test_samples,
						encoding_scheme=encoding_scheme, encoder=encoder)

	return train_set, val_set, test_set

def split_data_cv(data_file, datatype, problem_type, num_classes, excludeSeqID=False, 
					split_file=None, encoding_scheme='onehot', encoder=None,
					percent_val=0.15, percent_test=0.15, n_folds=5):
	"""Divide a datafile into training, val, test and 5 cross-val datasets.

	Takes in a datafile and specification of the data format and the machine
	learning problem, and returns PyTorch-compatible Dataset objects for
	the training, validation, test and cross-validation sets of the data. The
	user may optionally specify how the dataset should be split into these 
	subsets, as well as how protein sequences should be encoded as numeric 
	vectors.

	Parameters
	----------
	data_file : str
		Path to the datafile containing sequences and corresponding values
	datatype : str
		Format of the values within `data_file`. Should be 'sequence' if the
		`data_file` contains a single value per sequence, or 'residues' if
		it contains a value for each residue per sequence.
	problem_type : str
		The machine learning task to be addressed. Should be either 'regression'
		or 'classification'.
	excludeSeqID : bool, optional
		Flag that indicates how `data_file` is formatted. If False (default),
		then each line in the file should begin with a column containing a
		sequence ID. If True, then the datafile will not have this ID column,
		and will begin with the protein sequence.
	split_file : str, optional
		Path to a file containing information on how to divide the data into
		training, validation and test datasets. Default is None, which will
		cause the data to be divided randomly, with proportions based on
		`percent_val` and `percent_test`. If `split_file` is provided it must
		contain 3 lines in the file, corresponding to the training, validation
		and test sets. Each line should have whitespace-separated integer indices
		which correspond to lines in `data_file`.
	encoding_scheme : str, optional
		The method to be used for encoding protein sequences as numeric vectors.
		Currently 'onehot' and 'biophysics' are implemented (default is 'onehot').
	encoder: UserEncoder object, optional
		If encoding_scheme is 'user', encoder should be a UserEncoder object
		that can convert amino acid sequences to numeric vectors. If
		encoding_scheme is not 'user', use None.
	percent_val : float, optional
		If `split_file` is not provided, the fraction of the data that should be
		randomly assigned to the validation set. Should be in the range [0-1]
		(default is 0.15).
	percent_test : float, optional
		If `split_file` is not provided, the fraction of the data that should be
		randomly assigned to the test set. Should be in the range [0-1] (default
		is 0.15). The proportion of the training set will be calculated by the
		difference between 1 and the sum of `percent_val` and `percent_train`, so
		these should not sum to be greater than 1.
	n_folds : int, optional
		Number of folds for cross-validation (default is 5).

	Returns
	-------
	list of tuples of SequenceDataset objects
		a list of tuples of length `n_folds`. Each tuple contains the training
		and validation datasets for one of the cross-val folds.
	SequenceDataset object
		a dataset containing the training set sequences and values
	SequenceDataset object
		a dataset containing the validation set sequences and values
	SequenceDataset object
		a dataset containing the test set sequences and values
	"""

	data = parse_file(data_file, datatype, problem_type, num_classes, excludeSeqID=excludeSeqID)
	n_samples = len(data)

	# Initial step: split into training, val, and test sets
	if split_file == None:
		percent_train = 1 - percent_val - percent_test

		all_samples = np.arange(n_samples)
		training_samples, val_test_samples = vector_split(all_samples, percent_train)

		# Repeat procedure to split val and test sets
		val_test_fraction = percent_val / (percent_val + percent_test)
		val_samples, test_samples = vector_split(val_test_samples, val_test_fraction)

		# Generate datasets using these random partitions
		train_set = SequenceDataset(data=data, subset=training_samples, 
						encoding_scheme=encoding_scheme, encoder=encoder)
		val_set = SequenceDataset(data=data, subset=val_samples,
						encoding_scheme=encoding_scheme, encoder=encoder)
		test_set = SequenceDataset(data=data, subset=test_samples,
						encoding_scheme=encoding_scheme, encoder=encoder)

	# If provided, split datasets according to split_file
	else:
		training_samples, val_samples, test_samples = read_split_file(split_file)

		# Generate datasets using the provided partitions
		train_set = SequenceDataset(data=data, subset=training_samples,
						encoding_scheme=encoding_scheme, encoder=encoder)
		val_set = SequenceDataset(data=data, subset=val_samples,
						encoding_scheme=encoding_scheme, encoder=encoder)
		test_set = SequenceDataset(data=data, subset=test_samples,
						encoding_scheme=encoding_scheme, encoder=encoder)

	# Second step: combine train and val samples, and split evenly into n_folds
	cv_samples = np.append(training_samples, val_samples)
	np.random.shuffle(cv_samples)	# Shuffle train and val to avoid bias

	# Split into n_folds
	cv_samples = np.array_split(cv_samples, n_folds)

	# cv_sets will be a list of tuples: (fold_k_train_dataset, fold_k_test_dataset)
	cv_sets = []
	for i in range(len(cv_samples)):
		cv_test = cv_samples[i]
		cv_train = np.array([], dtype=int)
		for j in range(len(cv_samples)):
			if j != i:
				cv_train = np.append(cv_train, cv_samples[j])
		cv_train.sort()
		cv_test.sort()

		# Tuple of cross val train and test sets
		cv_sets.append( (SequenceDataset(data=data, subset=cv_train,
							encoding_scheme=encoding_scheme, encoder=encoder), 
						 SequenceDataset(data=data, subset=cv_test,
							encoding_scheme=encoding_scheme, encoder=encoder)) )

	return cv_sets, train_set, val_set, test_set



"""
This file contains code for conducting Bayesian optimization.

.............................................................................
idptools-parrot was developed by the Holehouse lab
     Original release ---- 2020

Question/comments/concerns? Raise an issue on github:
https://github.com/idptools/parrot

Licensed under the MIT license. 
"""

import math

import numpy as np
import GPy
import GPyOpt
from GPyOpt.methods import BayesianOptimization

from parrot import train_network
from parrot import brnn_architecture

class BayesianOptimizer(object):
	"""A class for conducting Bayesian Optimization on a PyTorch RNN

	Sets up and runs GPy Bayesian Optimization in order to choose the best-
	performing hyperparameters for a RNN for a given machine learning task. 
	Iteratively change learning rate, hidden vector size, and the number of layers
	in the network, then train and validating using 5-fold cross validation.

	Attributes
	----------
	cv_dataloaders : list of tuples of PyTorch DataLoader objects
		For each of the cross-val folds, a tuple containing a training set
		DataLoader and a validation set DataLoader.
	input_size : int
		Length of the amino acid encoding vectors
	n_epochs : int
		Number of epochs to train for each iteration of the algorithm
	n_classes : int
		Number of classes
	n_folds : int
		Number of cross-validation folds
	problem_type : str
		'classification' or 'regression'
	dtype : str
		'sequence' or 'residues'
	weights_file : str
		Path to which the network weights will be saved during training
	device : str
		'cpu' or 'cuda' depending on system hardware
	max_iterations : int
		Maximum number of iterations to perform the optimization procedure
	verbosity : int
		level of how descriptive the output to console message will be
	bds : list of dicts
		GPy-compatible bounds for each of the hyperparameters to be optimized

	Methods
	-------
	compute_cv_loss(hyperparameters)
		Compute the average cross-val loss for a given set of hyperparameters
	eval_cv_brnns(lr, nl, hs)
		Train and test a network with given parameters across all cross-val folds
	initial_search(x)
		Calculate loss and estimate noise for an initial set of hyperparameters
	optimize()
		Set up and run Bayesian Optimization on the BRNN using GPy
	"""

	def __init__(self, cv_dataloaders, input_size, n_epochs, n_classes, 
				dtype, weights_file, max_iterations, device, verbosity):
		"""
		Parameters
		----------
		cv_dataloaders : list of tuples of PyTorch DataLoader objects
			For each of the cross-val folds, a tuple containing a training set
			DataLoader and a validation set DataLoader.
		input_size : int
			Length of the amino acid encoding vectors
		n_epochs : int
			Number of epochs to train for each iteration of the algorithm
		n_classes : int
			Number of classes
		dtype : str
			'sequence' or 'residues'
		weights_file : str
			Path to which the network weights will be saved during training
		device : str
			'cpu' or 'cuda' depending on system hardware
		max_iterations : int
			Maximum number of iterations to perform the optimization procedure
		verbosity : int
			level of how descriptive the output to console message will be
		"""

		self.cv_loaders = cv_dataloaders
		self.input_size = input_size
		self.n_epochs = n_epochs
		self.n_folds = len(cv_dataloaders)
		self.n_classes = n_classes
		if n_classes > 1:
			self.problem_type = 'classification'
		else:
			self.problem_type = 'regression'

		self.dtype = dtype
		self.weights_file = weights_file
		self.max_iterations = max_iterations
		self.device = device
		self.verbosity = verbosity

		self.bds = [{'name': 'log_learning_rate', 'type': 'continuous', 'domain': (-5, 0)}, # 0.00001-1
					{'name': 'n_layers', 'type': 'discrete', 'domain': tuple(range(1, 11))}, # up to 10
					{'name': 'hidden_size', 'type': 'discrete', 'domain': tuple(range(1, 21))}] # up to 20

	def compute_cv_loss(self, hyperparameters):
		"""Compute the average cross-val loss for a given set of hyperparameters

		Given N sets of hyperparameters, determine the average cross-validation loss
		for BRNNs trained with these parameters.

		Parameters
		----------
		hyperparameters : numpy float array
			Each row corresponds to a set of hyperparameters, in the order:
			[log_learining_rate, n_layers, hidden_size]

		Returns
		-------
		numpy float array
			a Nx1 numpy array of the average cross-val loss 
			per set of input hyperparameters
		"""

		cv_outputs = np.zeros((len(hyperparameters), self.n_folds))

		for i in range(len(hyperparameters)):

			log_lr, nl, hs = hyperparameters[i]
			lr = 10**float(log_lr)
			nl = int(nl)
			hs = int(hs)

			# Train and validate network with these hyperparams using k-fold CV
			cv_outputs[i] = self.eval_cv_brnns(lr, nl, hs)
			avg = np.average(cv_outputs[i])

			if self.verbosity > 0:
				print('  %.6f	|     %2d       |         %2d           |    %.3f' % (lr, nl, hs, avg))

		outputs = np.average(cv_outputs, axis=1)
		return outputs


	def eval_cv_brnns(self, lr, nl, hs):
		"""Train and test a network with given parameters across all cross-val folds

		Parameters
		----------
		lr : float
			Learning rate of the network
		nl : int
			Number of hidden layers (for each direction) in the network
		hs : int
			Size of hidden vectors in the network

		Returns
		-------
		numpy float array
			the best validation loss from each fold of cross validation
		"""

		cv_losses = np.zeros(self.n_folds) - 1 # -1 so that it's obvious if something goes wrong

		for k in range(self.n_folds):
			if self.dtype == 'sequence':
				# Use a many-to-one architecture
				brnn_network = brnn_architecture.BRNN_MtO(self.input_size, hs, nl,
										self.n_classes, self.device).to(self.device)
			else:
				# Use a many-to-many architecture
				brnn_network = brnn_architecture.BRNN_MtM(self.input_size, hs, nl,
										self.n_classes, self.device).to(self.device)	

			# Train network with this set of hyperparameters
			train_losses, val_losses = train_network.train(brnn_network, self.cv_loaders[k][0],
										self.cv_loaders[k][1], self.dtype, self.problem_type,
										self.weights_file, stop_condition='iter', device=self.device,
										learn_rate=lr, n_epochs=self.n_epochs, verbosity=0)
			# Take best val loss
			best_val_loss = np.min(val_losses)
			cv_losses[k] = best_val_loss

		return cv_losses

	def initial_search(self, x):
		"""Calculate loss and estimate noise for an initial set of hyperparameters

		Parameters
		----------
		x : numpy array
			Array containing initial hyperparameters to test

		Returns
		-------
		numpy array
			Array containing the average losses of the input hyperparameters
		float
			The standard deviation of loss across cross-val folds for the
			input hyperparameters; an estimation of the training noise
		"""
		cv_outputs = np.zeros((len(x), self.n_folds))
		y = np.zeros((len(x), 1))

		for i in range(len(x)):

			log_lr, nl, hs = x[i]
			lr = 10**float(log_lr)
			nl = int(nl)
			hs = int(hs)

			# Train and validate network with these hyperparams using k-fold CV
			cv_outputs[i] = self.eval_cv_brnns(lr, nl, hs)

		# Calculate the avg and standard deviation of the losses
		for i in range(len(cv_outputs)):
			y[i] = np.average(cv_outputs[i])
		stddevs = np.std(cv_outputs, axis=1)
		avg_stddev = np.average(stddevs)

		return y, avg_stddev


	def optimize(self):
		"""Set up and run Bayesian Optimization on the BRNN using GPy

		Returns
		-------
		list
			the best hyperparameters are chosen by Bayesian Optimization. Returned
			in the order: [lr, nl, hs]
		"""

		# Initial hyperparameter search -- used to get noise estimate
		x_init = np.array([[-5.0, 5, 10], [-3.0, 5, 5], [-1.0, 8, 10], [-2.0, 10, 5], [-3.0, 3, 20]])
		y_init, noise = self.initial_search(x_init)

		if self.verbosity > 0:
			print("\nInitial search results:")
			print("lr\tnl\ths\toutput")
			for i in range(5):
				print("%.5f\t%2d\t%2d\t%.4f" % (10**x_init[i][0], x_init[i][1], x_init[i][2], y_init[i][0]))
			print("Noise estimate:", noise)				
			print('\n')
			print('Primary optimization:')
			print('--------------------\n')
			print('Learning rate   |   n_layers   |   hidden vector size |  avg CV loss  ')
			print('======================================================================')	

		optimizer = BayesianOptimization(f=self.compute_cv_loss, 
										 domain=self.bds,
										 model_type='GP',
										 acquisition_type='EI',
										 acquisition_jitter=0.05,
										 X=x_init,
										 Y=y_init,
										 noise_var=noise,
										 maximize=False)
		
		optimizer.run_optimization(max_iter=self.max_iterations)

		ins = optimizer.get_evaluations()[0]
		outs = optimizer.get_evaluations()[1].flatten()

		if self.verbosity > 0:
			print("\nThe optimal hyperparameters are:\nlr = %.6f\nnl = %d\nhs = %d" % 
						(10**optimizer.x_opt[0], optimizer.x_opt[1], optimizer.x_opt[2]))
			print()

		return optimizer.x_opt



"""
Unit and regression tests for the PARROT package. Note that these tests are by no means comprehensive.
For unexplained issues, please reach out on the GitHub page:
https://github.com/idptools/parrot/issues
"""

# Import package, test suite, and other packages as needed
import parrot
import pytest
import sys
import os
import pathlib
import shutil

def test_parrot_imported():
    """Sample test, will always pass so long as import statement worked"""
    assert "parrot" in sys.modules

def test_noarg_train_call():
	"""Calling parrot-train with no arguments should return an error"""
	os.chdir(pathlib.Path(__file__).parent.absolute())
	loc = shutil.which("parrot-train")
	script_descriptor = open(os.path.abspath(loc))
	script = script_descriptor.read()
	sys.argv = ["parrot-train"]

	with pytest.raises(SystemExit):
		exec(script)

	script_descriptor.close()

def test_train_invalid_dtype():
	"""Calling parrot-train with invalid datatype arg should return an error"""
	os.chdir(pathlib.Path(__file__).parent.absolute())
	loc = shutil.which("parrot-train")
	script_descriptor = open(os.path.abspath(loc))
	script = script_descriptor.read()
	sys.argv = ["parrot-train", "../data/seq_class_dataset.tsv", 
				"../data/output_network.pt", "-d", "gibberish", "-c", "4"]

	with pytest.raises(ValueError):
		exec(script)

	script_descriptor.close()

def test_train_invalid_numclasses():
	"""Calling parrot-train with invalid number of classes should return an error"""
	os.chdir(pathlib.Path(__file__).parent.absolute())
	loc = shutil.which("parrot-train")
	script_descriptor = open(os.path.abspath(loc))
	script = script_descriptor.read()
	sys.argv = ["parrot-train", "../data/seq_class_dataset.tsv", 
				"../data/output_network.pt", "-d", "sequence", "-c", "2"]

	with pytest.raises(IndexError):
		exec(script)

	script_descriptor.close()

def test_train_invalid_setFractions():
	"""Calling parrot-train with --setFractions not summing to 1 should return an error"""
	os.chdir(pathlib.Path(__file__).parent.absolute())
	loc = shutil.which("parrot-train")
	script_descriptor = open(os.path.abspath(loc))
	script = script_descriptor.read()
	sys.argv = ["parrot-train", "../data/seq_class_dataset.tsv", "../data/output_network.pt", 
				"-d", "sequence", "-c", "3", "--set-fractions", "0.8", "0.15", "0.15"]

	with pytest.raises(ValueError):
		exec(script)

	script_descriptor.close()

def test_noarg_optimize_call():
	"""Calling parrot-optimize with no arguments should return an error"""
	os.chdir(pathlib.Path(__file__).parent.absolute())
	loc = shutil.which("parrot-optimize")
	script_descriptor = open(os.path.abspath(loc))
	script = script_descriptor.read()
	sys.argv = ["parrot-optimize"]

	with pytest.raises(SystemExit):
		exec(script)

	script_descriptor.close()

def test_optimize_invalid_dtype():
	"""Calling parrot-optimize with invalid datatype arg should return an error"""
	os.chdir(pathlib.Path(__file__).parent.absolute())
	loc = shutil.which("parrot-optimize")
	script_descriptor = open(os.path.abspath(loc))
	script = script_descriptor.read()
	sys.argv = ["parrot-optimize", "../data/seq_class_dataset.tsv", 
				"../data/output_network.pt", "-d", "gibberish", "-c", "4"]

	with pytest.raises(ValueError):
		exec(script)

	script_descriptor.close()

def test_optimize_invalid_numclasses():
	"""Calling parrot-optimize with invalid number of classes should return an error"""
	os.chdir(pathlib.Path(__file__).parent.absolute())
	loc = shutil.which("parrot-optimize")
	script_descriptor = open(os.path.abspath(loc))
	script = script_descriptor.read()
	sys.argv = ["parrot-optimize", "../data/seq_class_dataset.tsv", 
				"../data/output_network.pt", "-d", "sequence", "-c", "2"]

	with pytest.raises(IndexError):
		exec(script)

	script_descriptor.close()

def test_split_data():
	"""Test that split_data() function is working as intended"""
	from parrot import process_input_data as pid

	data_file = os.path.abspath("../data/seq_class_dataset.tsv")
	train, val, test = pid.split_data(data_file, datatype='sequence', 
										problem_type='classification')

	assert (len(train) == 210) and (len(val) == 45) and (len(test) == 45) and (len(train[0]) == 3)

def test_split_data_cv():
	"""Test that split_data_cv() function is working as intended"""
	from parrot import process_input_data as pid

	data_file = os.path.abspath("../data/seq_class_dataset.tsv")
	cvs, train, val, test = pid.split_data_cv(data_file, datatype='sequence', 
										problem_type='classification')

	assert (len(train) == 210) and (len(val) == 45) and (len(test) == 45) and (len(train[0]) == 3) and (len(cvs) == 5) and (len(cvs[0]) == 2)




"""
File containing functions for encoding a string of amino acids into a numeric vector.

.............................................................................
parrot was developed by the Holehouse lab
     Original release ---- 2020

Question/comments/concerns? Raise an issue on github:
https://github.com/idptools/parrot

Licensed under the MIT license. 
"""

import sys
import os

import numpy as np
import torch

ONE_HOT = {'A':0, 'C':1, 'D':2, 'E':3, 'F':4, 'G':5, 'H':6, 'I':7, 'K':8, 'L':9,
		   'M':10,'N':11,'P':12,'Q':13,'R':14,'S':15,'T':16,'V':17,'W':18,'Y':19}

def one_hot(seq):
	"""Convert an amino acid sequence to a PyTorch tensor of one-hot vectors

	Each amino acid is represented by a length 20 vector with a single 1 and
	19 0's Inputing a sequence with a nono-canonical amino acid letter will
	cause the program to exit.

	E.g. Glutamic acid (E) is encoded: [0 0 0 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0]

	Parameters
	----------
	seq : str
		An uppercase sequence of amino acids (single letter code)

	Returns
	-------
	torch.IntTensor
		a PyTorch tensor representing the encoded sequence
	"""

	l = len(seq)
	m = np.zeros((l, 20))
	try:
		for i in range(l):
			m[i, ONE_HOT[seq[i]]] = 1
	except:
		error_str = 'Invalid amino acid detected: ' + seq[i]
		raise ValueError(error_str)
	return torch.from_numpy(m)

def rev_one_hot(seq_vectors):
	"""Decode a list of one-hot sequence vectors into amino acid sequences

	Parameters
	----------
	seq_vectors : list of numpy arrays
		A list containing sequence vectors

	Returns
	-------
	list
		Strings of amino acid sequences
	"""

	REV_ONE_HOT = 'ACDEFGHIKLMNPQRSTVWY'
	sequences = []

	for seq_vector in seq_vectors:
		seq = []
		for residue in seq_vector:
			seq.append(REV_ONE_HOT[np.argmax(residue)])
		sequences.append("".join(seq))

	return sequences


############  Biophysical scale encoding  #############

# Map each amino acid to a vector of biophysical properties
# 0: Hydrophobicity
# 1: Charge
# 2: pI
# 3: Molecular weight (g/mol)
# 4: Aromatic amino acid
# 5: Capable of hydrogen bonding
# 6: Side chain SASA (measured from ACE-XX-NME dipeptide)
# 7: Backbone SASA (measured from ACE-XX-NME dipeptide)
# 8: Free energy of solvation
BIOPHYSICS = {	'A':[ 1.8,  0,  6.0,  89.1, 0, 0,  75.8,  76.1,    1.9], 
				'C':[ 2.5,  0,  5.1, 121.2, 0, 0, 115.4,  67.9,   -1.2], 
				'D':[-3.5, -1,  2.8, 133.1, 0, 1, 130.3,  71.8, -107.3], 
				'E':[-3.5, -1,  3.2, 147.1, 0, 1, 161.8,  68.1, -107.3], 
				'F':[ 2.8,  0,  5.5, 165.2, 1, 0, 209.4,  66.0,   -0.8], 
				'G':[-0.4,  0,  6.0,  75.1, 0, 0,   0.0, 115.0,    0.0], 
				'H':[-3.2,  1,  7.6, 155.2, 0, 1, 180.8,  67.5,  -52.7], # Avg of HIP and HIE 
				'I':[ 4.5,  0,  6.0, 131.2, 0, 0, 172.7,  60.3,    2.2], 
				'K':[-3.9,  1,  9.7, 146.2, 0, 1, 205.9,  68.7, -100.9], 
				'L':[ 3.8,  0,  6.0, 131.2, 0, 0, 172.0,  64.5,    2.3],
		   		'M':[ 1.9,  0,  5.7, 149.2, 0, 0, 184.8,  67.8,   -1.4],
		   		'N':[-3.5,  0,  5.4, 132.1, 0, 1, 142.7,  66.8,   -9.7],
		   		'P':[-1.6,  0,  6.3, 115.1, 0, 0, 134.3,  55.8,    2.0],
		   		'Q':[-3.5,  0,  5.7, 146.2, 0, 1, 173.3,  66.6,   -9.4],
		   		'R':[-4.5,  1, 10.8, 174.2, 0, 1, 236.5,  66.7, -100.9],
		   		'S':[-0.8,  0,  5.7, 105.1, 0, 1,  95.9,  72.9,   -5.1],
		   		'T':[-0.7,  0,  5.6, 119.1, 0, 1, 130.9,  64.1,   -5.0],
		   		'V':[ 4.2,  0,  6.0, 117.1, 0, 0, 143.1,  61.7,    2.0],
		   		'W':[-0.9,  0,  5.9, 204.2, 1, 1, 254.6,  64.3,   -5.9],
		   		'Y':[-1.3,  0,  5.7, 181.2, 1, 1, 222.5,  71.9,   -6.1]
		   		}

def biophysics(seq):
	"""Convert an amino acid sequence to a PyTorch tensor with biophysical encoding

	Each amino acid is represented by a length 9 vector with each value representing
	a biophysical property. The nine encoded biophysical scales are Kyte-Doolittle
	hydrophobicity, charge, isoelectric point, molecular weight, aromaticity, 
	h-bonding ability, side chain solvent accessible surface area, backbone SASA, and 
	free energy of solvation. Inputing a sequence with a nono-canonical amino acid 
	letter will cause the program to exit.

	E.g. Glutamic acid (E) is: [-3.5, -1,  3.2, 147.1, 0, 1, 161.8,  68.1, -107.3]

	Parameters
	----------
	seq : str
		An uppercase sequence of amino acids (single letter code)

	Returns
	-------
	torch.FloatTensor
		a PyTorch tensor representing the encoded sequence
	"""
	l = len(seq)
	m = np.zeros((l, len(BIOPHYSICS['A'])))
	try:
		for i in range(l):
			m[i] = BIOPHYSICS[seq[i]]
	except:
		error_str = 'Invalid amino acid detected: ' + seq[i]
		raise ValueError(error_str)
	return torch.from_numpy(m)

def rev_biophysics(seq_vectors):
	"""Decode a list of biophysically-encoded sequence vectors into amino acid sequences

	Parameters
	----------
	seq_vectors : list of numpy arrays
		A list containing sequence vectors

	Returns
	-------
	list
		Strings of amino acid sequences
	"""

	REV_BIOPHYSICS = {}
	for key, value in BIOPHYSICS.items():
		REV_BIOPHYSICS[str(value[6])] = key

	sequences = []
	for seq_vector in seq_vectors:
		seq = []
		for residue in seq_vector:
			seq.append(REV_BIOPHYSICS[str(residue[6])])
		sequences.append("".join(seq))

	return sequences


################## User-specified encoding ####################

def parse_encode_file(file):
	"""Helper function to convert an encoding file into key:value dictionary"""

	with open(file) as f:
		lines = [x.strip().split() for x in f]

	l = len(lines[0]) - 1
	d = {}
	for line in lines:
		d[line[0]] = line[1:]

		if len(line) - 1 != l:
			raise ValueError('Vectors in encoding file do not have same length.')

	return d, l

class UserEncoder():
	"""User-specified amino acid-to-vector encoding scheme object

	Attributes
	----------
	encode_file : str
		A path to a file that describes the encoding scheme
	encode_dict : dict
		A dictionary that maps each amino acid to a numeric vector
	input_size : int
		The length of the encoding vector used for each amino acid
	"""

	def __init__(self, encode_file):
		"""
		Parameters
		----------
		encode_file : str
			A path to a file that describes the encoding scheme
		"""

		self.encode_file = os.path.abspath(encode_file)
		if not os.path.isfile(self.encode_file):
			raise FileNotFoundError('Encoding file does not exist.')

		self.encode_dict, self.input_size = parse_encode_file(self.encode_file)

	def __len__(self):
		"""Get length of encoding scheme"""

		return self.input_size

	def encode(self, seq):
		"""Convert an amino acid sequence into this encoding scheme

		Parameters
		----------
		seq : str
			An uppercase sequence of amino acids (single letter code)

		Returns
		-------
		torch.FloatTensor
			a PyTorch tensor representing the encoded sequence
		"""

		l = len(seq)
		m = np.zeros((l, self.input_size))

		try:
			for i in range(l):
				m[i] = self.encode_dict[seq[i]]
		except:
			error_str = 'Invalid amino acid detected: ' + seq[i]
			raise ValueError(error_str)
		return torch.from_numpy(m)

	def decode(self, seq_vectors):
		"""Converts a list of sequence vectors back to a list of protein sequences

		Parameters
		----------
		seq_vectors : list of numpy arrays
			A list containing sequence vectors

		Returns
		-------
		list
			Strings of amino acid sequences
		"""

		# Create a reverse of the encode_dict using hashing
		rev_dict = {}
		for key, value in self.encode_dict.items():
			value = np.array(value, dtype=np.float32)
			rev_dict[hash(tuple(value))] = key

		sequences = []
		for seq_vector in seq_vectors:
			seq = []
			for residue in seq_vector:
				seq.append(rev_dict[hash(tuple(residue))])
			sequences.append("".join(seq))

		return sequences


"""
Core training module of PARROT

.............................................................................
idptools-parrot was developed by the Holehouse lab
     Original release ---- 2020

Question/comments/concerns? Raise an issue on github:
https://github.com/idptools/parrot

Licensed under the MIT license. 
"""

import torch 
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader
import numpy as np

from parrot import brnn_plot
from parrot import encode_sequence

def train(network, train_loader, val_loader, datatype, problem_type, weights_file,
	stop_condition, device, learn_rate, n_epochs, verbosity=1):
	"""Train a BRNN and save the best performing network weights

	Train the network on a training set, and every epoch evaluate its performance on
	a validation set. Save the network weights that acheive the best performance on
	the validation set.

	User must specify the machine learning tast (`problem_type`) and the format of
	the data (`datatype`). Additionally, this function requires the learning rate
	hyperparameter and the number of epochs of training. The other hyperparameters, 
	number of hidden layers and hidden vector size, are implictly included on the 
	the provided network.

	The user may specify if they want to train the network for a set number of
	epochs or until an automatic stopping condition is reached with the argument
	`stop_condition`. Depending on the stopping condition used, the `n_epochs`
	argument will have a different role.

	Parameters
	----------
	network : PyTorch network object
		A BRNN network with the desired architecture
	train_loader : PyTorch DataLoader object
		A DataLoader containing the sequences and targets of the training set
	val_loader : PyTorch DataLoader object
		A DataLoader containing the sequences and targets of the validation set
	datatype : str
		The format of values in the dataset. Should be 'sequence' for datasets
		with a single value (or class label) per sequence, or 'residues' for
		datasets with values (or class labels) for every residue in a sequence.
	problem_type : str
		The machine learning task--should be either 'regression' or
		'classification'.
	weights_file : str
		A path to the location where the best_performing network weights will be
		saved
	stop_condition : str
		Determines when to conclude network training. If 'iter', then the network
		will train for `n_epochs` epochs, then stop. If 'auto' then the network
		will train for at least `n_epochs` epochs, then begin assessing whether
		performance has sufficiently stagnated. If the performance plateaus for
		`n_epochs` consecutive epochs, then training will stop.
	device : str
		Location of where training will take place--should be either 'cpu' or
		'cuda' (GPU). If available, training on GPU is typically much faster.
	learn_rate : float
		Initial learning rate of network training. The training process is
		controlled by the Adam optimization algorithm, so this learning rate
		will tend to decrease as training progresses.
	n_epochs : int
		Number of epochs to train for, or required to have stagnated performance
		for, depending on `stop_condition`.
	verbosity : int, optional
		The degree to which training updates are written to standard out (default
		is 1).

	Returns
	-------
	list
		A list of the average training set losses achieved at each epoch
	list
		A list of the average validation set losses achieved at each epoch
	"""

	# Max verbosity level is 2:
	if verbosity > 2:
		verbosity = 2

	# Set optimizer
	optimizer = torch.optim.Adam(network.parameters(), lr=learn_rate)

	# Set loss criteria
	if problem_type == 'regression':
		if datatype == 'residues':
			criterion = nn.MSELoss(reduction='sum')
		elif datatype == 'sequence':
			criterion = nn.L1Loss(reduction='sum')
	elif problem_type == 'classification':
		criterion = nn.CrossEntropyLoss(reduction='sum') 

	network = network.float()
	total_step = len(train_loader)
	min_val_loss = np.inf
	avg_train_losses = []
	avg_val_losses = []

	if stop_condition == 'auto':
		min_epochs = n_epochs
		# Set to some arbitrarily large number of iterations -- will stop automatically
		n_epochs = 20000000
		last_decrease = 0

	# Train the model - evaluate performance on val set every epoch
	end_training = False
	for epoch in range(n_epochs):	# Main loop

		# Initialize training and testing loss for epoch
		train_loss = 0
		val_loss = 0

		# Iterate over batches
		for i, (names, vectors, targets) in enumerate(train_loader):		# CHANGEME
			vectors = vectors.to(device)
			targets = targets.to(device)
		
			# Forward pass
			outputs = network(vectors.float())

			if problem_type == 'regression':
				loss = criterion(outputs, targets.float())
			else:
				if datatype == 'residues':
					outputs = outputs.permute(0, 2, 1)
				loss = criterion(outputs, targets.long())

			train_loss += loss.data.item()
			
			# Backward and optimize
			optimizer.zero_grad()
			loss.backward()
			optimizer.step()

		for names, vectors, targets in val_loader:
			vectors = vectors.to(device)
			targets = targets.to(device)

			# Forward pass
			outputs = network(vectors.float())
			if problem_type == 'regression':
				loss = criterion(outputs, targets.float())
			else:
				if datatype == 'residues':
					outputs = outputs.permute(0, 2, 1)
				loss = criterion(outputs, targets.long())

			# Increment val loss
			val_loss += loss.data.item()

		# Avg loss:
		train_loss /= len(train_loader.dataset)
		val_loss /= len(val_loader.dataset)

		signif_decrease = True
		if stop_condition == 'auto' and epoch > min_epochs - 1:
			# Check to see if loss has stopped decreasing
			last_epochs_loss = avg_val_losses[-min_epochs:]

			for loss in last_epochs_loss:
				if val_loss >= loss*0.995:
					signif_decrease = False

			# If network performance has plateaued over the last range of epochs, end training
			if not signif_decrease and epoch - last_decrease > min_epochs:
				end_training = True

		# Only save updated weights to memory if they improve val set performance
		if val_loss < min_val_loss:
			min_val_loss = val_loss 	# Reset min_val_loss
			last_decrease = epoch
			torch.save(network.state_dict(), weights_file)	# Save model

		# Append losses to lists
		avg_train_losses.append(train_loss)
		avg_val_losses.append(val_loss)

		if verbosity == 2:
			print('Epoch %d\tLoss %.4f' % (epoch, val_loss))
		elif epoch % 5 == 0 and verbosity == 1:
			print('Epoch %d\tLoss %.4f' % (epoch, val_loss))

		# This is placed here to ensure that the best network, even if the performance
		# improvement is marginal, is saved.
		if end_training:
			break

	# Return loss per epoch so that they can be plotted
	return avg_train_losses, avg_val_losses


def test_labeled_data(network, test_loader, datatype, problem_type, 
						weights_file, num_classes, device):
	"""Test a trained BRNN on labeled sequences

	Using the saved weights of a trained network, run a set of sequences through
	the network and evaluate the performancd. Return the average loss per
	sequence and plot the results. Testing a network on previously-unseen data 
	provides a useful estimate of how generalizeable the network's performance is.

	Parameters
	----------
	network : PyTorch network object
		A BRNN network with the desired architecture
	test_loader : PyTorch DataLoader object
		A DataLoader containing the sequences and targets of the test set
	datatype : str
		The format of values in the dataset. Should be 'sequence' for datasets
		with a single value (or class label) per sequence, or 'residues' for
		datasets with values (or class labels) for every residue in a sequence.
	problem_type : str
		The machine learning task--should be either 'regression' or
		'classification'.
	weights_file : str
		A path to the location of the best_performing network weights
	num_classes: int
		Number of data classes. If regression task, put 1.
	device : str
		Location of where testing will take place--should be either 'cpu' or
		'cuda' (GPU). If available, training on GPU is typically much faster.

	Returns
	-------
	float
		The average loss across the entire test set
	list of lists
		Details of the output predictions for each of the sequences in the test set. Each
		inner list represents a sample in the test set, with the format: [sequence_vector,
		true_value, predicted_value, sequence_ID]
	"""

	# Load network weights
	network.load_state_dict(torch.load(weights_file))

	# Get output directory for images
	network_filename = weights_file.split('/')[-1]
	output_dir = weights_file[:-len(network_filename)]

	# Set loss criteria
	if problem_type == 'regression':
		if datatype == 'residues':
			criterion = nn.MSELoss(reduction='sum')
		elif datatype == 'sequence':
			criterion = nn.L1Loss(reduction='sum')
	elif problem_type == 'classification':
		if datatype == 'residues':
			criterion = nn.CrossEntropyLoss(reduction='sum')
		elif datatype == 'sequence':
			criterion = nn.CrossEntropyLoss()

	test_loss = 0
	all_targets = []
	all_outputs = []
	predictions = []
	for names, vectors, targets in test_loader: 	# batch size of 1
		all_targets.append(targets)

		vectors = vectors.to(device)
		targets = targets.to(device)

		# Forward pass
		outputs = network(vectors.float())
		if problem_type == 'regression':
			loss = criterion(outputs, targets.float())
		else:
			if datatype == 'residues':
				outputs = outputs.permute(0, 2, 1)
			loss = criterion(outputs, targets.long())
	
		test_loss += loss.data.item() # Increment test loss
		all_outputs.append(outputs.detach())

		# Add to list as: [seq_vector, true value, predicted value, name]
		predictions.append( [vectors[0].cpu().numpy(), targets.cpu().numpy()[0], outputs.cpu().detach().numpy(), names[0]] )

	# Plot 'accuracy' depending on the problem type and datatype
	if problem_type == 'regression':
		if datatype == 'residues':
			brnn_plot.residue_regression_scatterplot(all_targets, all_outputs, output_dir=output_dir)

			# Format predictions
			for i in range(len(predictions)):
				predictions[i][2] = predictions[i][2].flatten()
				predictions[i][1] = predictions[i][1].flatten()

		elif datatype == 'sequence':
			brnn_plot.sequence_regression_scatterplot(all_targets, all_outputs, output_dir=output_dir)

			# Format predictions
			for i in range(len(predictions)):
				predictions[i][2] = predictions[i][2][0][0]
				predictions[i][1] = predictions[i][1][0]

	elif problem_type == 'classification':

		if datatype == 'residues':
			brnn_plot.res_confusion_matrix(all_targets, all_outputs, num_classes, output_dir=output_dir)

			# Format predictions and assign class predictions
			for i in range(len(predictions)):
				pred_values = []
				for j in range(len(predictions[i][2])):
					pred_values = np.argmax(predictions[i][2], axis=1)[0]
				predictions[i][2] = np.array(pred_values, dtype=np.int)

		elif datatype == 'sequence':
			brnn_plot.confusion_matrix(all_targets, all_outputs, num_classes, output_dir=output_dir)

			# Format predictions and assign class predictions
			for i in range(len(predictions)):
				pred_value = np.argmax(predictions[i][2])
				predictions[i][2] = int(pred_value)
				predictions[i][1] = predictions[i][1]

	return test_loss / len(test_loader.dataset), predictions

def test_unlabeled_data(network, sequences, device, encoding_scheme='onehot', encoder=None):
	"""Test a trained BRNN on unlabeled sequences

	Use a trained network to make predictions on previously-unseen data.

	** Note: Unlike the previous functions, `network` here must have pre-loaded
	weights. **

	Parameters
	----------
	network : PyTorch network object
		A BRNN network with the desired architecture and pre-loaded weights
	sequences : list
		A list of amino acid sequences to test using the network
	device : str
		Location of where testing will take place--should be either 'cpu' or
		'cuda' (GPU). If available, training on GPU is typically much faster.
	encoding_scheme : str, optional
		How amino acid sequences are to be encoded as numeric vectors. Currently,
		'onehot','biophysics' and 'user' are the implemented options.
	encoder: UserEncoder object, optional
		If encoding_scheme is 'user', encoder should be a UserEncoder object
		that can convert amino acid sequences to numeric vectors. If
		encoding_scheme is not 'user', use None.

	Returns
	-------
	dict
		A dictionary containing predictions mapped to sequences
	"""

	pred_dict = {}
	for seq in sequences:
		if encoding_scheme == 'onehot':
			seq_vector = encode_sequence.one_hot(seq)
		elif encoding_scheme == 'biophysics':
			seq_vector = encode_sequence.biophysics(seq)
		elif encoding_scheme == 'user':
			seq_vector = encoder.encode(seq)

		seq_vector = seq_vector.view(1, len(seq_vector), -1)

		# Forward pass
		outputs = network(seq_vector.float()).detach().numpy()
		pred_dict[seq] = outputs

	return pred_dict



"""
Plot training results for regression and classification tasks on both 
sequence-mapped and residue-mapped data.

.............................................................................
idptools-parrot was developed by the Holehouse lab
     Original release ---- 2020

Question/comments/concerns? Raise an issue on github:
https://github.com/idptools/parrot

Licensed under the MIT license. 
"""

import numpy as np
import torch
import itertools
from scipy.stats import linregress
import matplotlib.pyplot as plt
import seaborn as sn
import pandas as pd

from parrot import encode_sequence

def training_loss(train_loss, val_loss, output_dir=''):
	"""Plot training and validation loss per epoch

	Figure is not displayed, but saved to file in current directory with the name
	'train_test.png'.

	Parameters
	----------
	train_loss : list
		training loss across each epoch
	val_loss : list
		validation loss across each epoch
	output_dir : str, optional
		directory to which the plot will be saved (default is current directory)
	"""

	fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(6,6))
	props = dict(boxstyle='round', facecolor='gainsboro', alpha=0.5)

	num_epochs = len(train_loss)

	# Loss per epoch
	training_loss, = ax.plot(np.arange(1, num_epochs+1), train_loss, label='Train')
	validation_loss, = ax.plot(np.arange(1, num_epochs+1), val_loss, label='Val')
	ax.set_xlabel("Epoch")
	ax.set_ylabel("Avg loss")
	ax.set_title("Training and testing loss per epoch")
	ax.legend(handles=[training_loss, validation_loss], fontsize=14, 
					facecolor='gainsboro', edgecolor='slategray')


	if num_epochs < 21:
		ax.set_xticks(np.arange(2, num_epochs+1, 2))
	elif num_epochs < 66:
		ax.set_xticks(np.arange(5, num_epochs+1, 5))
	elif num_epochs < 151:
		ax.set_xticks(np.arange(10, num_epochs+1, 10))
	else:
		ax.set_xticks(np.arange(50, num_epochs+1, 50))

	plt.savefig(output_dir + 'train_test.png')
	plt.clf()


def sequence_regression_scatterplot(true, predicted, output_dir=''):
	"""Create a scatterplot for a sequence-mapped values regression problem

	Figure is displayed to console if possible and saved to file in current 
	directory with the name 'seq_scatter.png'.

	Parameters
	----------
	true : list of PyTorch FloatTensors
		A list where each item is a [1 x 1] tensor with the true regression value
		of a particular sequence
	predicted : list of PyTorch FloatTensors
		A list where each item is a [1 x 1] tensor with the regression prediction
		for a particular sequence
	output_dir : str, optional
		directory to which the plot will be saved (default is current directory)
	"""

	true_list = []
	pred_list = []

	for item in true:
		true_list.append(item.cpu().numpy()[0][0])
	for item in predicted:
		pred_list.append(item.cpu().numpy()[0][0])

	plt.scatter(true_list, pred_list)
	edge_vals = [0.9*min(min(true_list), min(pred_list)), 
				 1.1*max(max(true_list), max(pred_list))]
	plt.xlim(edge_vals)
	plt.ylim(edge_vals)
	plt.plot(edge_vals, edge_vals, 'k--')
	plt.xlabel('True')
	plt.ylabel('Predicted')
	slope, intercept, r_value, p_value, std_err = linregress(true_list, pred_list)
	plt.title('Testing accuracy: R^2=%.3f' % (r_value**2))
	plt.savefig(output_dir + 'seq_scatter.png')
	plt.show()


def residue_regression_scatterplot(true, predicted, output_dir=''):
	"""Create a scatterplot for a residue-mapped values regression problem

	Each sequence is plotted with a unique marker-color combination, up to 70
	different sequences.

	Figure is displayed to console if possible and saved to file in current 
	directory with the name 'res_scatter.png'.

	Parameters
	----------
	true : list of PyTorch FloatTensors
		A list where each item is a [1 x len(sequence)] tensor with the true
		regression values of each residue in a sequence
	predicted : list of PyTorch FloatTensors
		A list where each item is a [1 x len(sequence)] tensor with the 
		regression predictions for each residue in a sequence
	output_dir : str, optional
		directory to which the plot will be saved (default is current directory)
	"""

	true_list = []
	pred_list = []

	marker = itertools.cycle(('>', '+', '.', 'o', '*', 'v', 'D')) 

	for item in true:
		single_frag = item.cpu().numpy()[0].flatten()
		true_list.append(list(single_frag))
	for item in predicted:
		single_frag = item.cpu().numpy()[0].flatten()
		pred_list.append(list(single_frag))

	for i in range(len(true_list)):
		plt.scatter(true_list[i], pred_list[i], s=6, marker=next(marker))

	plt.figure(1)

	left, right = plt.xlim()
	bottom, top = plt.ylim()
	edge_vals = [min(left, bottom), max(right, top)]
	plt.xlim(edge_vals)
	plt.ylim(edge_vals)
	plt.plot(edge_vals, edge_vals, 'k--')
	plt.xlabel('True')
	plt.ylabel('Predicted')
	slope, intercept, r_value, p_value, std_err = linregress(sum(true_list, []), sum(pred_list, []))
	plt.title('Testing accuracy: R^2=%.3f' % (r_value**2))
	plt.savefig(output_dir + 'res_scatter.png')
	plt.show()


def confusion_matrix(true_classes, predicted_classes, num_classes, output_dir=''):
	"""Create a confusion matrix for a sequence classification problem

	Figure is displayed to console if possible and saved to file in current 
	directory with the name 'seq_CM.png'.

	Parameters
	----------
	true_classes : list of PyTorch IntTensors
		A list where each item is a [1 x 1] tensor with the true class label of a
		particular sequence
	predicted_classes : list of PyTorch FloatTensors
		A list where each item is a [1 x num_classes] tensor prediction of the
		class label for a particular sequence
	num_classes : int
		Number of distinct data classes
	output_dir : str, optional
		directory to which the plot will be saved (default is current directory)
	"""

	cm = np.zeros((num_classes, num_classes))
	for i in range(len(true_classes)):
		cm[true_classes[i][0], np.argmax(predicted_classes[i][0].cpu().numpy())] += 1

	df_cm = pd.DataFrame(cm, range(num_classes), range(num_classes))
	sn.set(font_scale=1.4) # for label size
	sn.heatmap(df_cm, cmap='Blues', annot=True, annot_kws={"size": 16}) # font size
	plt.ylabel('True labels')
	plt.xlabel('Predicted labels')
	plt.title('Test set confusion matrix')
	plt.tight_layout()
	plt.savefig(output_dir + 'seq_CM.png')
	plt.show()


def res_confusion_matrix(true_classes, predicted_classes, num_classes, output_dir=''):
	"""Create a confusion matrix for a residue classification problem

	Figure is displayed to console if possible and saved to file in current 
	directory with the name 'res_CM.png'.

	Parameters
	----------
	true_classes : list of PyTorch IntTensors
		A list where each item is a [1 x len(sequence)] tensor with the true class
		label of the residues in a particular sequence
	predicted_classes : list of PyTorch FloatTensors
		A list where each item is a [1 x num_classes x len(sequence)] tensor
		with predictions of the class label for each residue in a particular
		sequence
	num_classes : int
		Number of distinct data classes
	output_dir : str, optional
		directory to which the plot will be saved (default is current directory)
	"""

	true_list = []
	pred_list = []

	for item in true_classes:
		single_frag = list(item[0].cpu().numpy().flatten())
		true_list = true_list + single_frag

	for item in predicted_classes:
		single_frag = item[0].permute(1, 0).cpu().numpy()

		for residue in single_frag:
			pred_list.append(np.argmax(residue))

	cm = np.zeros((num_classes, num_classes))
	for i in range(len(true_list)):
		cm[true_list[i], pred_list[i]] += 1

	df_cm = pd.DataFrame(cm, range(num_classes), range(num_classes))
	sn.set(font_scale=1.4) # for label size
	sn.heatmap(df_cm, cmap='Blues', annot=True, annot_kws={"size": 16}) # font size
	plt.ylabel('True labels')
	plt.xlabel('Predicted labels')
	plt.title('Test set confusion matrix')
	plt.tight_layout()
	plt.savefig(output_dir + 'res_CM.png')
	plt.show()

def output_predictions_to_file(sequence_data, excludeSeqID, encoding_scheme, encoder=None,
																			output_dir=''):
	"""Output sequences, their true values, and their predicted values to a file

	Used on the output of the test_unlabeled_data() function in the train_network module in
	order to detail the performance of the trained network on the test set. Produces the
	file "test_set_predictions.tsv" in output_dir. Each pair of lines in this tsvfile
	corresponds to a particular test set sequence, with the first containing the true data
	values, and the second line having the predicted data values.

	Parameters
	----------
	sequence_data : list of lists
		Details of the output predictions for each of the sequences in the test set. Each
		inner list represents a sample in the test set, with the format: [sequence_vector,
		true_value, predicted_value, sequence_ID]
	excludeSeqID : bool
		Boolean indicating whether or not each line in `tsvfile` has a sequence ID
		(default is False)		
	encoding_scheme : str
		Description of how an amino acid sequence should be encoded as a numeric 
		vector. Providing a string other than 'onehot', 'biophysics', or 'user' 
		will produce unintended consequences.
	encoder: UserEncoder object, optional
		If encoding_scheme is 'user', encoder should be a UserEncoder object
		that can convert amino acid sequences to numeric vectors. If
		encoding_scheme is not 'user', use None.
	"""

	seq_vectors = []
	true_vals = []
	pred_vals = []
	names = []
	count = 0
	for sequence in sequence_data:
		seq_vector, true_val, pred_val, name = sequence
		seq_vectors.append(seq_vector)
		true_vals.append(true_val)
		pred_vals.append(pred_val)

		if excludeSeqID:
			names.append('test' + str(count))
			count+=1
		else:
			names.append(name)

	# Decode the sequence vectors
	if encoding_scheme == 'onehot':
		sequences = encode_sequence.rev_one_hot(seq_vectors)
	elif encoding_scheme == 'biophysics':
		sequences = encode_sequence.rev_biophysics(seq_vectors)
	else:
		sequences = encoder.decode(seq_vectors)

	# Write to file
	filename = 'test_set_predictions.tsv'

	with open(output_dir + filename, 'w') as tsvfile:
		for i in range(len(names)):

			# Adjust formatting for residues or sequence data
			if isinstance(true_vals[i], np.ndarray):
				true_vals_format = ' '.join(true_vals[i].astype(str))
				pred_vals_format = ' '.join(pred_vals[i].astype(str))
			else:
				true_vals_format = true_vals[i]
				pred_vals_format = pred_vals[i]

			'''
			Format:
			NAME_TRUE SEQUENCE TRUE_VALUE(S)
			NAME_PRED SEQUENCE PRED_VALUE(S)
			'''
			output_str = "%s_TRUE %s %s\n" % (names[i], sequences[i], true_vals_format)
			output_str = output_str + "%s_PRED %s %s\n" % (names[i], sequences[i], pred_vals_format)
	
			tsvfile.write(output_str)
